module BgUtils
	class Skybox
		def initialize(filename)
			# filename is supposed to lead to a composed texture like below :
			#   X
			#  XXXX
			#   X

			texture = Gosu::Image.new(filename, :retro=>true)
			@tile_size = texture.width / 4
			texture = nil # we don't need it anymore
			@textures = {
				:y_pos => Gosu::Image.new(filename, :rect => [@tile_size, 0, @tile_size, @tile_size]),
				:x_neg => Gosu::Image.new(filename, :rect => [0, @tile_size, @tile_size, @tile_size]),
				:z_neg => Gosu::Image.new(filename, :rect => [@tile_size, @tile_size, @tile_size, @tile_size]),
				:x_pos => Gosu::Image.new(filename, :rect => [2 * @tile_size, @tile_size, @tile_size, @tile_size]),
				:z_pos => Gosu::Image.new(filename, :rect => [3 * @tile_size, @tile_size, @tile_size, @tile_size]),
				:y_neg => Gosu::Image.new(filename, :rect => [@tile_size, 2 * @tile_size, @tile_size, @tile_size])
			}
			self.gen_list
		end

		def switch_texture(face)
			tex = @textures[face].gl_tex_info
			glBindTexture(GL_TEXTURE_2D, tex.tex_name)
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
			return tex
		end

		def gen_list
			@display_list = glGenLists(1)
    	    glNewList(@display_list, GL_COMPILE)
				glDepthMask(GL_FALSE)
				tex = switch_texture(:z_neg)
				glBegin(GL_QUADS)
					glTexCoord2d(tex.left, tex.top); 		glVertex3f(-1, 1, -1)
					glTexCoord2d(tex.left, tex.bottom); 	glVertex3f(-1, -1, -1)
					glTexCoord2d(tex.right, tex.bottom); 	glVertex3f(1, -1, -1)
					glTexCoord2d(tex.right, tex.top); 		glVertex3f(1, 1, -1)
				glEnd
				tex = switch_texture(:z_pos)
				glBindTexture(GL_TEXTURE_2D, tex.tex_name)
				glBegin(GL_QUADS)
					glTexCoord2d(tex.left, tex.top); 		glVertex3f(1, 1, 1)
					glTexCoord2d(tex.left, tex.bottom); 	glVertex3f(1, -1, 1)
					glTexCoord2d(tex.right, tex.bottom); 	glVertex3f(-1, -1, 1)
					glTexCoord2d(tex.right, tex.top); 		glVertex3f(-1, 1, 1)
				glEnd
				tex = switch_texture(:x_neg)
				glBindTexture(GL_TEXTURE_2D, tex.tex_name)
				glBegin(GL_QUADS)
					glTexCoord2d(tex.left, tex.top); 		glVertex3f(-1, 1, 1)
					glTexCoord2d(tex.left, tex.bottom); 	glVertex3f(-1, -1, 1)
					glTexCoord2d(tex.right, tex.bottom); 	glVertex3f(-1, -1, -1)
					glTexCoord2d(tex.right, tex.top); 		glVertex3f(-1, 1, -1)
				glEnd
				tex = switch_texture(:x_pos)
				glBindTexture(GL_TEXTURE_2D, tex.tex_name)
				glBegin(GL_QUADS)
					glTexCoord2d(tex.left, tex.top); 		glVertex3f(1, 1, -1)
					glTexCoord2d(tex.left, tex.bottom); 	glVertex3f(1, -1, -1)
					glTexCoord2d(tex.right, tex.bottom); 	glVertex3f(1, -1, 1)
					glTexCoord2d(tex.right, tex.top); 		glVertex3f(1, 1, 1)
				glEnd
				tex = switch_texture(:y_pos)
				glBindTexture(GL_TEXTURE_2D, tex.tex_name)
				glBegin(GL_QUADS)
					glTexCoord2d(tex.left, tex.top); 		glVertex3f(-1, 1, 1)
					glTexCoord2d(tex.left, tex.bottom); 	glVertex3f(-1, 1, -1)
					glTexCoord2d(tex.right, tex.bottom); 	glVertex3f(1, 1, -1)
					glTexCoord2d(tex.right, tex.top); 		glVertex3f(1, 1, 1)
				glEnd
				tex = switch_texture(:y_neg)
				glBindTexture(GL_TEXTURE_2D, tex.tex_name)
				glBegin(GL_QUADS)
					glTexCoord2d(tex.left, tex.top); 		glVertex3f(-1, -1, -1)
					glTexCoord2d(tex.left, tex.bottom); 	glVertex3f(-1, -1, 1)
					glTexCoord2d(tex.right, tex.bottom); 	glVertex3f(1, -1, 1)
					glTexCoord2d(tex.right, tex.top); 		glVertex3f(1, -1, -1)
				glEnd
				glDepthMask(GL_TRUE)	        	
        	glEndList
		end

		def draw(x = 0, y = 0, z = 0, rot_pitch = 0, rot_yaw = 0)		
			glPushMatrix
				glTranslatef(x, y, z)
				glRotatef(rot_pitch, 1, 0, 0)
				glRotatef(rot_yaw, 0, 1, 0)
				glCallList(@display_list)
			glPopMatrix
			
		end
	end

	class EmptyImageSource
		private
		def initialize(columns, rows, gosu_color)
			@columns, @rows = columns, rows
			@gosu_color = gosu_color
		end

		def columns; @columns; end
		def rows; @rows; end

		def to_blob
			color = [@gosu_color.red, @gosu_color.green, @gosu_color.blue, @gosu_color.alpha]
			color.pack("C*") * @columns * @rows
		end
	end

	public
	def self.get_pixel(gosu_image, x, y)
		if x < 0 or x >= gosu_image.width or y < 0 or y >= gosu_image.height
			return nil
		else
			result = gosu_image.to_blob[(y * gosu_image.width + x) * 4, 4].unpack("C*")
			return Gosu::Color.new(result[3], result[0], result[1], result[2])
		end		
	end

	def self.set_pixel(gosu_image, x, y, color, brush_size = 1)
		one_pixel_image = Gosu::Image.new(EmptyImageSource.new(brush_size.floor, brush_size.floor, color))
		gosu_image.insert(one_pixel_image, x.floor, y.floor)
	end

	class TiledImage
		attr_reader :width, :height, :scale_x, :scale_y
		def initialize(gosu_image, width, height, scale_x = 1.0, scale_y = 1.0, color = 0xff_ffffff, mode = :default, options = {:retro => true})
			options[:tileable] = true
			if gosu_image.is_a?(Gosu::Image) or gosu_image.is_a?(String)
				@image = (gosu_image.is_a?(Gosu::Image)) ? gosu_image : Gosu::Image.new(gosu_image, options)
			else
				raise("Wrong parameter 0 for 'Image' : \ngiven : #{gosu_image}\nshould be either a Gosu::Image instance or a string that contains path and filename")
			end

			@width, @height, @scale_x, @scale_y = width, height, scale_x, scale_y
			@color, @mode = color, mode
		end

		def draw(x, y, z, scale_x = 1, scale_y = 1)
			@drawing_macro ||= create_macro
			@drawing_macro.draw(x, y, z, scale_x, scale_y)
		end

		def draw_rot(x, y, z, angle, center_x = 0.5, center_y = 0.5, scale_x = 1, scale_y = 1)
			@drawing_macro ||= create_macro
			@drawing_macro.draw_rot(x, y, z, angle, center_x, center_y, scale_x, scale_y)
		end

		def draw_as_quad(x1, y1, x2, y2, x3, y3, x4, y4, z)
			@drawing_macro ||= create_macro
			c = Gosu::Color::WHITE
			@drawing_macro.draw_as_quad(x1, y1, c, x2, y2, c, x3, y3, c, x4, y4, c, z)
		end

		private
		def create_macro
			Gosu::record(@width, @height) do
				tile_width = @image.width * @scale_x
				tile_height = @image.height * @scale_y
				c = @color

				# filling with as many full tiles as possible
				for tile_x in 0...(@width / tile_width).floor
					for tile_y in 0...(@height / tile_height).floor
						@image.draw(tile_x * tile_width, tile_y * tile_height, 0, @scale_x, @scale_y, c, @mode)
					end
				end			

				# filling extra width with part of the tile if needed
				if (extra_width = @width % tile_width) > 0
					subimage_width = (extra_width / @scale_x).floor
					subimage = @image.subimage(0, 0, subimage_width, @image.height)
					for tile_y in 0...(@height / tile_height).floor
						subimage.draw_as_quad(
							@width - extra_width, tile_y * tile_height, c, 
							@width, tile_y * tile_height, c,
							@width - extra_width, (tile_y + 1) * tile_height, c,
							@width, (tile_y + 1) * tile_height, c, 
							0, @mode
						)
					end
				end		

				# filling extra height with part of the tile if needed
				if (extra_height = @height % tile_height) > 0
					subimage_height = (extra_height / @scale_y).floor
					subimage = @image.subimage(0, 0, @image.width, subimage_height)
					for tile_x in 0...(@width / tile_width).floor
						subimage.draw_as_quad(
							tile_x * tile_width, @height - extra_height, c, 
							(tile_x + 1) * tile_width, @height - extra_height, c,
							tile_x * tile_width, @height, c,
							(tile_x + 1) * tile_width, @height, c, 
							0, @mode
						)
					end
				end	

				# if there were extra pixels in both axis, we have to draw a bottom right corner
				if (extra_width > 0 and extra_height > 0)
					subimage_width = (extra_width / @scale_x).floor
					subimage_height = (extra_height / @scale_y).floor
					subimage = @image.subimage(0, 0, subimage_width, subimage_height)

					subimage.draw_as_quad(
						@width - extra_width, @height - extra_height, c,
						@width, @height - extra_height, c,
						@width - extra_width, @height, c,
						@width, @height, c,
						0, @mode
					)
				end
			end
		end
	end
end